#include <cassert>
#include <cstdlib> // 'std::srand', etc.
#include <ctime> // 'std::time_t'

#include <filesystem> // 'std::filesystem::path'...
#include <future> // 'std::future', 'std::promise'...
#include <fstream> // 'std::filebuf', 'std::ifstream'...
#include <iostream>
#include <list>
#include <mutex>
#include <string>
#include <system_error> // 'std::error_code'...
#include <thread>

#include "Promise.h"

std::list<std::launch> Image::_Start; // Benchmarking only...
std::list<std::launch> Image::_End; // Benchmarking only...
std::mutex Image::_Mutex; // Benchmarking only...

Image::Image(std::string&& file_name) :_file_name(file_name) {}

std::string Image::process(const std::launch& launch) noexcept(false) {
	/*std::ofstream readme;
	readme.open("Readme.txt");
	readme << _file_name + " just processed...\n";
	readme.close();*/

	std::string start = launch == std::launch::deferred ? "Deferred" :
		launch == (std::launch::async | std::launch::deferred) ? "Default" :
		launch == std::launch::async ? "Async" : "?";
	start += " (start)";
	{
		const std::lock_guard<std::mutex> critical_section(_Mutex);
		_Start.push_back(launch);
		for (std::launch launch : _Start) {
			start += " < ";
			start += launch == std::launch::deferred ? "Deferred" :
				launch == (std::launch::async | std::launch::deferred) ? "Default" :
				launch == std::launch::async ? "Async" : "?";
		}
	} // 'mutex' is released at 'critical_section' destruction time

	std::size_t size = 0;
	try {
		std::ifstream ifstream(_file_name, std::ios::binary | std::ios::in);
		assert(ifstream.is_open());
		std::filebuf* buffer = ifstream.rdbuf();
		assert(buffer);
		// Get file size by setting internal position to relative position and returns new position:
		size = buffer->pubseekoff(0, ifstream.end, ifstream.in);
		// Set internal position to absolute position (very beginning):
		std::size_t position = buffer->pubseekpos(0, ifstream.in);
		assert(position == 0);
		// Allocate 'memory' to contain file data
		char* memory = new char[size];
		// Fill in 'memory' with file data:
		buffer->sgetn(memory, size);
		// 'ifstream' is automatically closed at destruction time:
		// ifstream.close();
		// Write file data to stdout:
		std::cout.write(memory, size);
		delete[] memory;
	}
	catch (std::ifstream::failure & failure) {
		std::cerr << failure.what() << std::endl;
		// Attempt to partially correct...
		throw;
	}

	std::string end = launch == std::launch::deferred ? "Deferred" :
		launch == (std::launch::async | std::launch::deferred) ? "Default" :
		launch == std::launch::async ? "Async" : "?";
	end += " (end)";
	{
		const std::lock_guard<std::mutex> critical_section(_Mutex);
		_End.push_back(launch);
		for (std::launch launch : _Start) {
			end += " < ";
			end += launch == std::launch::deferred ? "Deferred" :
				launch == (std::launch::async | std::launch::deferred) ? "Default" :
				launch == std::launch::async ? "Async" : "?";
		}
	} // 'mutex' is released at 'critical_section' destruction time

	return start + " *** " + end;
}

Promise::Promise() {
	std::time_t t;
	// See also 'std::default_random_engine' at https://en.cppreference.com/w/cpp/numeric/random
	std::srand(static_cast<unsigned>(std::time(&t))); // Setup random suite...
}

std::string Promise::get_Resource() {
	std::future<std::string> f = _p.get_future(); // Blocking!
	std::string resource_name = f.get();
	std::cout << "'get_Resource' ends for " << resource_name << std::endl;
	return resource_name;
}

void Promise::request_Resource(std::string&& resource_name) {
	// This is a job that takes time to compute some value...
	int wait_for = std::abs(std::rand() % 5000 + 1); // Between 1 and 5000...
	std::cout << "Wait for " << wait_for << " ms about: " << resource_name << std::endl;
	std::this_thread::sleep_for(std::chrono::milliseconds(wait_for));
	_p.set_value(resource_name); // Job ends... '_p.get_future();' is unblocked...
}

int main() {
	/*
	Promise franck, barbier;
	franck.request_Resource("Franck");
	barbier.request_Resource("Barbier");
	std::cout << franck.get_Resource() + barbier.get_Resource() << std::endl;
	*/

	// C++17:
	//std::filesystem::path path("/"); // Unix style adapts to Windows, i.e., "C:\\"
	//path /= "Users";
	//path /= "franc";
	//path /= "Desktop";
	//path /= "Franck.jpg";
	//std::cout << "exists(): " << std::boolalpha << std::filesystem::exists(path) << "\n" // 'true'
	//	<< "root_name(): " << path.root_name() << "\n" // ""
	//	<< "root_path(): " << path.root_path() << "\n" // "/"
	//	<< "relative_path(): " << path.relative_path() << "\n" // "Users\\franc\\Desktop\\Franck.jpg"
	//	<< "parent_path(): " << path.parent_path() << "\n" // "/Users\\franc\\Desktop"
	//	<< "filename(): " << path.filename() << "\n" // "Franck.jpg"
	//	<< "stem(): " << path.stem() << "\n" // "Franck"
	//	<< "extension(): " << path.extension() << "\n"; // ".jpg"
	//if (std::filesystem::exists(path) && std::filesystem::is_regular_file(path)) {
	//	std::error_code error_code;
	//	std::uintmax_t size = std::filesystem::file_size(path, error_code);
	//	if (size != static_cast<std::uintmax_t>(-1))
	//		std::cout << "size: " << size << "\n"; // '6176'
	//	else
	//		std::cout << "size (error): " << error_code.category().message(error_code.value()) << "\n";
	//}

	Image image("Franck.jpg");
	// Passed functions are each executed in a separate thread:
	std::future<std::string> deferred_f = std::async(std::launch::deferred, std::bind(&Image::process, image, std::launch::deferred));
	std::future<std::string> default_f = std::async(std::bind(&Image::process, image, std::launch::async | std::launch::deferred));
	std::future<std::string> async_f = std::async(std::launch::async, std::bind(&Image::process, image, std::launch::async));

	std::string async_s = async_f.get(); // Blocking...
	std::string deferred_s = deferred_f.get(); // Blocking...
	std::string default_s = default_f.get(); // Blocking...

	std::cout << std::endl << std::endl;
	std::cout << deferred_s << std::endl;
	std::cout << default_s << std::endl;
	std::cout << async_s << std::endl;
	
	return 0;
}